//
//  token.cpp
//  Lily
//
//  Created by KMY on 2015/05/08.
//  Copyright (c) 2015年 KMY. All rights reserved.
//

#include "token.h"
#include "script.h"
#include "message.h"

Token::Token(Script* script) {
	scriptObj = script;
	tokenLchar = 0;
}

Token::~Token() {
	delete tokenLchar;
}

void Token::analysis() {
	
	// スクリプトを取得
	const lchar* scriptLchar = scriptObj->getScriptIntoLchar();
	const int scriptLen = Character::lstrlen(scriptLchar);
	scriptObj->setLineNumber(1);
	
	// 解析結果保存用
	tokenLchar = new lchar[scriptLen*2];
	
	// 各フラグ
	bool isAddToken = false;			// 現在のトークンに追加するか
	bool isEndToken = false;			// 現在の文字がトークンの最後の文字であるか
	bool isNewToken = false;			// 現在の文字が新しいトークンの最初の文字であるか
	lchar* tokenType = &tokenLchar[0];	// トークンの種類保存先
	lchar* beforeTokenType = 0;			// 前のトークン
	lchar newTokenType = TOKEN_UNKNOWN;	// isNewTokenを指定した時の次のトークンの種類
	lchar* tokenCurLchar = &tokenLchar[1];	// 現在処理中の文字が格納されるべき場所
	bool isString = false;
	bool isStringKakko = false;			// 「」
	bool isStringNkakko = false;		// 『』
	bool isStringDquote = false;		// ""
	bool isStringSquote = false;		// ''
	bool isSpace = false;
	bool isBeforeSpace = false;			// 前の文字がスペースやタブ
	bool isJava = false;				// Java部分
	bool isJavaParam = false;			// Javaの引数挿入部分
	bool isComment = true;
	bool isCommentLine = true;			// 単数行コメント
	bool isCommentArea = true;			// 複数行コメント
	
	// 最初のトークンの初期化
	*tokenType = TOKEN_UNKNOWN;
	
	// １文字ずつ反復
	for (int i=0; i<scriptLen; i++) {
		
		lchar c = scriptLchar[i];			// 現在の文字
		
		// トークン関連のフラグ
		isAddToken = true;
		isNewToken = false;
		isEndToken = false;
		
		// ----------------------------------------------------
		// 通常のプログラム中の処理（文字列、Java出力部分などでない）
		// ----------------------------------------------------
		
		if(!isString && !isJava) {
		
			// ----------------------------------------------------
			// 空白処理
			// ----------------------------------------------------

			isBeforeSpace = isSpace;
			isSpace = false;
			if(Character::equal(c, " ") || Character::equal(c, "　") || Character::equal(c, "\t")) {
				isSpace = true;
				
				// スペースの大きさを確認
				int len = 0;
				if (Character::equal(c, " ")) {
					len = 1;
				} else if (Character::equal(c, "　")) {
					len = 2;
				} else {
					len = -1;
				}
				
				// スペース直後でないときの対応
				if (!isBeforeSpace) {

					// インデントであるか？
					if (!beforeTokenType || *beforeTokenType == TOKEN_YN) {
						newTokenType = TOKEN_INDENT;
						isNewToken = true;
						c = len > 0 ? len : _TABLEN;
					}
					
					// インデントでなければ無視
					else {
						isAddToken = false;
						newTokenType = TOKEN_UNKNOWN;
						isNewToken = true;
					}
				}
				
				// 連続スペースのときの対応
				else {
					
					// インデントが連続した場合
					if (*tokenType == TOKEN_INDENT) {
						isAddToken = false;
						
						// タブの長さを設定
						if (len < 0) {
							len = _TABLEN - *(tokenCurLchar-1) % _TABLEN;
						}
						
						*(tokenCurLchar-1) += len;
					}
					
					// インデントでなければ無視
					else {
						isAddToken = false;
					}
				}
			}
			
			// ----------------------------------------------------
			// 空白でないすべての文字の処理
			// ----------------------------------------------------
			
			else {
				
				// スペースの終端であるか判定
				if (isBeforeSpace) {
					isEndToken = true;
				}
				
				// 文字列の開始を判定
				if(Character::equal(c, "「")) {
					isString = isStringKakko = true;
				} else if(Character::equal(c, "『")) {
					isString = isStringNkakko = true;
				} else if(Character::equal(c, "\"")) {
					isString = isStringDquote = true;
				} else if(Character::equal(c, "\'")) {
					isString = isStringSquote = true;
				}
				// 文字列の開始ならかっこを含めない
				if (isString) {
					newTokenType = Token::TOKEN_STR;
					isAddToken = false;
					isNewToken = true;
				}
			
				// ----------------------------------------------------
				// 助詞
				// ----------------------------------------------------
				
				else {
					int joshiType = Token::JOSHI_UNDEF;
					if (Character::equal(c, "と")) {
						joshiType = JOSHI_TO;
					}
					else if (Character::equal(c, "を")) {
						joshiType = JOSHI_WO;
					}
					else if (Character::equal(c, "は")) {
						// とは
						if (beforeTokenType && *beforeTokenType == TOKEN_JOSHI && *(beforeTokenType+1) == JOSHI_TO) {
							tokenType = beforeTokenType;
							tokenCurLchar = tokenType + 1;
							joshiType = JOSHI_TOHA;
						}
						else
							joshiType = JOSHI_HA;
					}
					else if (Character::equal(*(tokenCurLchar-1), "なら")) {
						joshiType = JOSHI_NARA;
						
						// すでに助詞前の文章と同一トークンになってしまってるため、NULL文字追加などで切り離す
						tokenType = tokenCurLchar;
						tokenCurLchar = tokenType + 1;
						*(tokenType-1) = 0;
					}
					else if (Character::equal(c, "ば")) {
						// ならば
						if (beforeTokenType && *beforeTokenType == TOKEN_JOSHI && *(beforeTokenType+1) == JOSHI_NARA) {
							tokenType = beforeTokenType;
							tokenCurLchar = tokenType + 1;
							joshiType = JOSHI_NARABA;
						}
					}
					// 助詞であれば助詞の種類を保存してトークンを終了
					if (joshiType != Token::JOSHI_UNDEF) {
						newTokenType = Token::TOKEN_JOSHI;
						isNewToken = true;
						c = joshiType;
						isEndToken = true;
					}
					
					// ----------------------------------------------------
					// 命令の終了／区切り、改行、演算子、その他各記号
					// ----------------------------------------------------
					
					else {
						
						// 命令終端や改行
						if (Character::equal(c, "。")) {
							isNewToken = true;
							newTokenType = Token::TOKEN_ENDINST;
							isEndToken = true;
						}
						else if (Character::equal(c, "、")) {
							isNewToken = true;
							newTokenType = Token::TOKEN_ENDPUNC;
							isEndToken = true;
						}
						else if (Character::equal(c, "\n")) {
							// \r\n
							if (beforeTokenType && *beforeTokenType == TOKEN_YN && Character::equal(*(beforeTokenType+1), "\r")) {
								tokenType = beforeTokenType;
								tokenCurLchar = tokenType + 1;
								isEndToken = true;
							}
							else {
								isNewToken = true;
								newTokenType = Token::TOKEN_YN;
								isEndToken = true;
							}
						}
						else if (Character::equal(c, "\r")) {
							isNewToken = true;
							newTokenType = Token::TOKEN_YN;
							isEndToken = true;
						}
						
						// 演算子、記号
						else if (Character::equal(c, "<") || (Character::equal(c, "＜"))) {
							isNewToken = true;
							newTokenType = Token::TOKEN_SYMBOL;
							c = SYMBOL_DAINARI;
							isEndToken = true;
						} else if (Character::equal(c, ">") || (Character::equal(c, "＞"))) {
							isNewToken = true;
							newTokenType = Token::TOKEN_SYMBOL;
							c = SYMBOL_SYOUNARI;
							isEndToken = true;
						} else if (Character::equal(c, "(") || (Character::equal(c, "（"))) {
							isNewToken = true;
							newTokenType = Token::TOKEN_SYMBOL;
							c = SYMBOL_MKAKKOS;
							isEndToken = true;
						} else if (Character::equal(c, ")") || (Character::equal(c, "）"))) {
							isNewToken = true;
							newTokenType = Token::TOKEN_SYMBOL;
							c = SYMBOL_MKAKKOE;
							isEndToken = true;
						} else if (Character::equal(c, "?") || (Character::equal(c, "？"))) {
							// <?
							if (beforeTokenType && *beforeTokenType == TOKEN_SYMBOL && *(beforeTokenType+1) == SYMBOL_DAINARI) {
								tokenType = beforeTokenType;
								tokenCurLchar = tokenType + 1;
								c = SYMBOL_JAVA_START;
								isEndToken = true;
								isJava = true;
							}
							else {
								isNewToken = true;
								newTokenType = Token::TOKEN_SYMBOL;
								c = SYMBOL_QUES;
								isEndToken = true;
							}
						}
						
						// その他特殊なブロック
						else if (Character::equal(c, "▲")) {
							isNewToken = true;
							newTokenType = Token::TOKEN_S_MACRO;
							isEndToken = true;
						}
					}
				}
			}
		}
		
		// ----------------------------------------------------
		// Java中であるときの処理
		// ----------------------------------------------------
		
		else if (isJava) {
			
			*tokenType = TOKEN_JAVA_SRC;
			
			// 引数挿入部分であるか
			if (isJavaParam) {
				
				*tokenType = TOKEN_JAVA_PARAM;
				
				if (Character::equal(c, "｝")) {
					isEndToken = true;
					isAddToken = false;
					isJavaParam = false;
				}
			}
			
			// 引数挿入部分の開始であるか
			else if (Character::equal(c, "｛")) {
				isNewToken = true;
				isAddToken = false;
				isJavaParam = true;
				newTokenType = TOKEN_JAVA_PARAM;
			}
			
			// ?>
			else if (Character::equal(c, ">") || Character::equal(c, "＞")) {
				lchar bc = *(tokenCurLchar-1);
				if (Character::equal(bc, "?") || Character::equal(bc, "？")) {
					tokenCurLchar -= 2;
					newTokenType = TOKEN_SYMBOL;
					isNewToken = true;
					c = SYMBOL_JAVA_END;
					isEndToken = true;
					isJava = false;
				}
			}
		}
		
		// ----------------------------------------------------
		// 文字列リテラル中であるときの処理
		// ----------------------------------------------------
		
		else if (isString) {
			
			// 文字列の終端を判定
			if (isStringKakko) {
				if (Character::equal(c, "」")) {
					isString = isStringKakko = false;
				}
			}
			else if (isStringNkakko) {
				if (Character::equal(c, "』")) {
					isString = isStringNkakko = false;
				}
			}
			else if (isStringDquote) {
				if (Character::equal(c, "\"")) {
					isString = isStringDquote = false;
				}
			}
			else if (isStringSquote) {
				if (Character::equal(c, "\'")) {
					isString = isStringSquote = false;
				}
			}
			// 文字列の終端ならかっこを含めない
			if (!isString) {
				isAddToken = false;
				isEndToken = true;
			}
		}
		
		// ----------------------------------------------------
		// 指定されたフラグにしたがってトークン作成処理
		// ----------------------------------------------------
		
		// 新しいトークンを生成
		if (isNewToken) {
			
			// 古いトークンに文字が存在しない時、今の領域をそのまま使う
			if (tokenCurLchar == tokenType+1) {
				*tokenType = newTokenType;
			}
			
			// 古いトークンに文字が存在すれば、新しいトークンの領域を取得
			else {
				beforeTokenType = tokenType;
				*tokenCurLchar = 0;			// NULL文字
				tokenType = tokenCurLchar+1;
				tokenCurLchar = tokenType+1;
				*tokenType = newTokenType;
			}
		}
		
		// 文字をトークンに追加
		if (isAddToken) {
			*tokenCurLchar = c;
			tokenCurLchar++;
		}
		
		// 古いトークンの終了
		if (isEndToken) {
			
			// 古いトークンに文字が存在しない時、今の領域をそのまま使う
			if (tokenCurLchar == tokenType+1) {
				*tokenType = TOKEN_UNKNOWN;
			}
			
			// 古いトークンに文字が存在すれば、新しいトークンの領域を取得
			else {
				beforeTokenType = tokenType;
				*tokenCurLchar = 0;			// NULL文字
				tokenType = tokenCurLchar+1;
				tokenCurLchar = tokenType+1;
				*tokenType = TOKEN_UNKNOWN;
			}
		}
	}
	
	// ファイル終端の処理
	if (tokenCurLchar != tokenType+1) {
		*tokenCurLchar = 0;			// NULL文字
		tokenType = tokenCurLchar+1;
		tokenCurLchar = tokenType+1;
	}
	*tokenType = TOKEN_ENDFILE;
	*tokenCurLchar = 0;
	
	// 全トークンの表示
	//tokenTestAll(tokenLchar);
}

const lchar* Token::getTokenLchar() {
	return tokenLchar;
}

const lchar* Token::nextToken(const lchar *p) {
	p++;				// 型の種類が0（VT_UNKNOWN）である可能性があるため
	while (*p) p++;
	return p+1;
}

void Token::tokenTest(const lchar *p) {
	if (*p == TOKEN_JOSHI) {
		printf(" -t:助詞　　\t%s\n",
			   (
				*(p+1) == JOSHI_UNDEF ? "未定義助詞" :
				*(p+1) == JOSHI_TO ? "と" :
				*(p+1) == JOSHI_HA ? "は" :
				*(p+1) == JOSHI_WO ? "を" :
				*(p+1) == JOSHI_TOHA ? "とは" :
				*(p+1) == JOSHI_NARA ? "なら" :
				*(p+1) == JOSHI_NARABA ? "ならば" :
				"不明な助詞"
				));
	} else if (*p == TOKEN_SYMBOL) {
			printf(" -t:記号　　\t%s\n",
				   (
					*(p+1) == SYMBOL_SYOUNARI ? "＞" :
					*(p+1) == SYMBOL_DAINARI ? "＜" :
					*(p+1) == SYMBOL_QUES ? "？" :
					*(p+1) == SYMBOL_MKAKKOS ? "（" :
					*(p+1) == SYMBOL_MKAKKOE ? "）" :
					*(p+1) == SYMBOL_JAVA_START ? "Java部分開始" :
					*(p+1) == SYMBOL_JAVA_END ? "Java部分終了" :
					"不明な記号"
					));

	} else if (*p == TOKEN_YN) {
		printf(" -t:改行\n");
	} else if (*p == TOKEN_INDENT) {
		printf(" -t:字下げ　\t%d\n", *(p+1));
	} else {
		char* str;
		Character::toCharn(p+1, &str);
		printf(" -t:%s\t%s\n",
			   (
				*p == TOKEN_CONST ? "定数　　" :
				*p == TOKEN_ENDFILE ? "ファイル終端" :
				*p == TOKEN_ENDINST ? "命令終端" :
				*p == TOKEN_ENDPUNC ? "命令区切" :
				*p == TOKEN_FUNC ? "関数　　" :
				*p == TOKEN_JOSHI ? "助詞　　" :
				*p == TOKEN_NUM ? "数値　　" :
				*p == TOKEN_STATICSTR ? "静的文字列" :
				*p == TOKEN_STR ? "文字列　" :
				*p == TOKEN_UNKNOWN ? "不明　　" :
				*p == TOKEN_VAR ? "変数　　" :
				*p == TOKEN_SPACE ? "空白　　" :
				*p == TOKEN_INDENT ? "字下げ　" :
				*p == TOKEN_SYMBOL ? "記号　　" :
				*p == TOKEN_JAVA_SRC ? "Java　　" :
				*p == TOKEN_JAVA_PARAM ? "Java埋込" :
				*p == TOKEN_S_MACRO ? "マクロ始" :
				*p == TOKEN_S_FUNCTION ? "関数始　" :
				*p == TOKEN_YN ? "改行　　" : "未定義　"
			   ),
			   str
			   );
		delete str;
	}
}

void Token::tokenTestAll(const lchar *p) {
	
	int i = 0;
	while (*(p+i) != TOKEN_ENDFILE) {

		// 内容表示
		tokenTest(p+i);

		// 次のNULL文字まで進める
		do {
			i++;
		} while (*(p+i));
		i++;
	}
	tokenTest(p+i);
}

